import java.util.ArrayList;
import java.util.Random;
import Jama.Matrix;

public class Architecture implements Cloneable {

	private int no_of_components;
	private int no_of_processes;
	private double[] component_rel;
	private int[] replicas;
	private int[] maxReplicas;
	private boolean[][] deployment;
	private double[] component_rel_eff;
	private double[][] opProfile;
	private double[] component_mem;
	private double[] component_cpu;
	private int[] processUsage;
	private double processMemRequirement;
	private double processCPURequirement;
	private double totalMemAvailable;
	private double totalCPUAvailable;
	private double utility;
	private int trial;

	public Architecture(int no_of_components, int no_of_processes) {
		this.no_of_components = no_of_components;
		this.no_of_processes = no_of_processes;
		this.component_rel = new double[no_of_components];
		this.replicas = new int[no_of_components];
		this.maxReplicas = new int[no_of_components];
		this.component_rel_eff = new double[no_of_components];
		this.deployment = new boolean[no_of_components][no_of_processes];
		this.component_mem = new double[no_of_components];
		this.component_cpu = new double[no_of_components];
		this.processUsage = new int[no_of_processes];
	}

	public void setComponentReliability(int componentId, double rel)
			throws Exception {
		// Check if such component exists
		if (componentId >= this.no_of_components || componentId < 0)
			throw new Exception("No Such Component Exists");
		// If it exists, set the reliability
		this.component_rel[componentId] = rel;
	}

	public void setComponentMemUsage(int componentId, double mem)
			throws Exception {
		// Check if such component exists
		if (componentId >= this.no_of_components || componentId < 0)
			throw new Exception("No Such Component Exists");
		// If it exists, set the memory
		this.component_mem[componentId] = mem;
	}

	public void setComponentCPUUsage(int componentId, double cpu)
			throws Exception {
		// Check if such component exists
		if (componentId >= this.no_of_components || componentId < 0)
			throw new Exception("No Such Component Exists");
		// If it exists, set the CPU usage
		this.component_cpu[componentId] = cpu;
	}

	public void setMaxReplicas(int componentId, int maxReplicas) {
		this.maxReplicas[componentId] = maxReplicas;
	}

	public void setOperationalProfile(double[][] opProfile) {
		this.opProfile = opProfile;
	}

	public void setProcessCPUReq(double cpuReq) {
		this.processCPURequirement = cpuReq;
	}

	public void setProcessMemReq(double memReq) {
		this.processMemRequirement = memReq;
	}

	public void setAvailableCPU(double cpu) {
		this.totalCPUAvailable = cpu;
	}

	public void setAvailableMem(double mem) {
		this.totalMemAvailable = mem;
	}

	public void deployComponent(int componentId, int processId) {
		if (canDeployComponent(componentId, processId)) {
			// if it is already on the same process, no changes to be done
			int currentProcessId = -1;
			boolean isCurrentlyDeployed = false;
			// find current processId
			for (int j = 0; j < no_of_processes; j++) {
				if (deployment[componentId][j] == true) {
					currentProcessId = j;
					isCurrentlyDeployed = true;
					break;
				}
			}

			// if currently deployed, then undeploy it from its current process
			if (isCurrentlyDeployed) {
				deployment[componentId][currentProcessId] = false;
				processUsage[currentProcessId]--;
			}

			// deploy the component
			deployment[componentId][processId] = true;
			processUsage[processId]++;
		} else
			throw new DeploymentNotAllowedException();
	}

	public boolean canDeployComponent(int componentId, int processId) {
		// Check if such component exists
		if (componentId >= this.no_of_components || componentId < 0)
			throw new RuntimeException("No Such Component Exists");
		// Check if such process exists
		if (processId >= this.no_of_processes || processId < 0)
			throw new RuntimeException("No Such Process Exists");
		// Check if any other component in the target process is replicated
		boolean canDeploy = false;
		if (processUsage[processId] == 0)
			// if target process is vacant, then no problem
			canDeploy = true;
		else {
			// if it is in use make sure the component, and the target process's
			// components have no replicas
			boolean thisCompOK = false;
			boolean otherCompsOK = true;
			if (replicas[componentId] == 0)
				thisCompOK = true;
			for (int i = 0; i < no_of_components; i++) {
				if (deployment[i][processId] == true && replicas[i] > 0) {
					otherCompsOK = false;
					break;
				}
			}
			if (thisCompOK && otherCompsOK)
				canDeploy = true;
		}
		return canDeploy;
	}

	public boolean canAddReplica(int componentId) {
		boolean maxOK = false;
		boolean deploymentOK = false;
		if (!hasReachedMaxReplicas(componentId))
			maxOK = true;
		if (isInItsOwnProcess(componentId))
			deploymentOK = true;
		return (maxOK && deploymentOK);
	}

	public boolean canRemoveReplica(int componentId) {
		return this.replicas[componentId] > 0;
	}

	public void addReplica(int componentId) {
		if (canAddReplica(componentId))
			this.replicas[componentId]++;
		else
			throw new ReplicationNotAllowedException();
	}

	public void removeReplica(int componentId) {
		if (canRemoveReplica(componentId))
			this.replicas[componentId]--;
	}

	public boolean hasReachedMaxReplicas(int componentId) {
		return this.replicas[componentId] == this.maxReplicas[componentId];
	}

	public void setReplicas(int componentId, int replicas) {
		if (replicas > 0) {
			if (replicas > this.maxReplicas[componentId])
				throw new RuntimeException(
						"Replication not allowed: Exceeds maximum allowed replicas");
			if (!isInItsOwnProcess(componentId))
				throw new RuntimeException(
						"Replication not allowed: Component is sharing a process");
		}
		this.replicas[componentId] = replicas;

	}

	public void optimizeByExhaustiveSearch(double reliablityReq) {

		System.out.println("Begining to search exhaustively ...");

		// reset deployment
		// for (int i = 0; i < no_of_components; i++) {
		// setReplicas(i, 0);
		// deployComponent(i, 0);
		// }

		// begin with zeroth component and search exhaustively
		Architecture bestArch = optimizeByExhaustiveSearchForComponent(
				reliablityReq, (Architecture) this.clone(), 0);
		System.out
				.println("The best architecture found after exhaustive search...");
		bestArch.show();

	}

	public Architecture optimizeByExhaustiveSearchForComponent(
			double reliablityReq, Architecture currentBestArch, int componentId) {

		for (int j = 0; j < no_of_processes; j++) {

			if (canDeployComponent(componentId, j)) {
				deployComponent(componentId, j);

				if (componentId < (no_of_components - 1)) {

					currentBestArch = this
							.optimizeByExhaustiveSearchForComponent(
									reliablityReq, currentBestArch,
									componentId + 1);

				} else {
					trial++;
					System.out.println("Trial : " + trial);
					show();

					if ((computeSystemReliability() >= reliablityReq)
							&& (getUtility(this) >= getUtility(currentBestArch))) {
						currentBestArch = (Architecture) this.clone();
					} else if (computeSystemReliability() >= reliablityReq) {
						currentBestArch = (Architecture) this.clone();
					}
				}

				setReplicas(componentId, 0);
				while (canAddReplica(componentId)) {

					addReplica(componentId);

					if (componentId < (no_of_components - 1)) {

						currentBestArch = this
								.optimizeByExhaustiveSearchForComponent(
										reliablityReq, currentBestArch,
										componentId + 1);

					} else {
						trial++;
						System.out.println("Trial : " + trial);
						show();

						if ((computeSystemReliability() >= reliablityReq)
								&& (getUtility(this) >= getUtility(currentBestArch))) {
							currentBestArch = (Architecture) this.clone();
						} else if (computeSystemReliability() >= reliablityReq) {
							currentBestArch = (Architecture) this.clone();
						}
					}
				}
				setReplicas(componentId, 0);

			}

		}

		return currentBestArch;

	}

	public void optimizeArchitecture_New(double[][] opProfile,
			double reliablityReq) throws Exception {
		// Two possible paths here: Predicted system reliability is 1) less than
		// requirement 2) higher than requirement
		if (computeSystemReliability() < reliablityReq) {
			ArrayList explored = new ArrayList(0);
			while (explored.size() != component_rel.length) {
				int current_component = getNextMostUnReliableComponent(explored);
				System.out.println("Consider Component : " + current_component
						+ ".....");
				// Heuristic 1: deploy the next-most unreliable component to a
				// vacant process
				// if it is not in its own process, deploy to a new process
				if (!isInItsOwnProcess(current_component)) {
					System.out.println("Apply heuristic 1 to component");
					int vacantProcess = getVacantProcess();
					if (vacantProcess != -1
							&& canDeployComponent(current_component,
									vacantProcess)) {
						deployComponent(current_component, vacantProcess);
						System.out.println("Deployed component to process "
								+ vacantProcess);
						// if requirement is met, then star random search
						if (computeSystemReliability() >= reliablityReq) {
							System.out
									.println("New architecture meets reliability constraint");
							show();
							performRandomNeighborhoodSearch(reliablityReq);
							return;
						} else {
							System.out
									.println("New architecture does not meet reliability constraint");
						}
					} else if (vacantProcess != -1) {
						System.out
								.println("No processes available for component to be deployed");
					} else {
						System.out
								.println("Component cannot be deployed to a different process");
					}
				} else {
					System.out
							.println("Cannot apply heuristic 1 to component since it is already deployed as seperate process");
				}
				// Heuristic 2: go on replicating the component to a new process
				if (canAddReplica(current_component)) {
					System.out.println("Apply heuristic 2 to component");
					boolean keepGoing = true;
					while (keepGoing) {
						addReplica(current_component);
						System.out.println("Replicated component");
						if (computeSystemReliability() >= reliablityReq) {
							System.out
									.println("New architecture meets reliability constraint");
							show();
							performRandomNeighborhoodSearch(reliablityReq);
							return;
						} else {
							System.out
									.println("New architecture does not meet reliability constraint");
						}
						if (!canAddReplica(current_component)) {
							keepGoing = false;
						}
					}
				} else {
					System.out
							.println("Cannot apply heuristic 2 to component as no more replicas can be added");
				}
				// Since both heuristics have been applied to the component, add
				// it to the "explored" set
				explored.add(current_component);
			}
			// if we are here, then no solution is found
			System.out
					.println("No architecure could be found to meet the reliability requirement. The best architecture is :");
			this.show();
			// ////////////////////Ehsan
			performRandomNeighborhoodSearch(reliablityReq);
			return;
			// ///////////////////Ehsan
		} else {
			System.out.println("No optimization required");
		}
	}

	public void performRandomNeighborhoodSearch(double reliablityReq)
			throws Exception {
		System.out
				.println("Performing random neighborhood search to further optimize architecture");
		// consider the current architecture to be the starting point of the
		// random search
		Architecture currentBestArch = (Architecture) this.clone();
		double currentBestUtility = getUtility(currentBestArch);
		// begin random neighborhood search
		Random rand = new Random();
		for (int i = 1; i < 10; i++) {
			System.out.println("Iteration " + i + ".....");
			try {
				int randomComponent = rand.nextInt(no_of_components);
				System.out.println("Random component selected : "
						+ randomComponent);
				int randomProcess = rand.nextInt(no_of_processes);
				System.out
						.println("Random process selected : " + randomProcess);
				// If possible re-deploy it randomly
				if (currentBestArch.canDeployComponent(randomComponent,
						randomProcess)) {
					System.out.println("Redeployed component to process");
					Architecture candidateArch = (Architecture) currentBestArch
							.clone();
					candidateArch.deployComponent(randomComponent,
							randomProcess);
					if (candidateArch.computeSystemReliability() >= reliablityReq) {
						if (candidateArch.getUtility(candidateArch) > currentBestUtility) {
							System.out
									.println("New architecture has better utility than previous candidate");
							currentBestArch = candidateArch;
							currentBestUtility = candidateArch
									.getUtility(candidateArch);
							currentBestArch.show();
						} else {
							System.out
									.println("New architecture does not have better utility than previous candidate");
						}
					} else {
						System.out
								.println("New architecture does not meet reliability requirement");
					}
				} else {
					System.out
							.println("Component cannot be redeployed to process");
				}

				if (currentBestArch.canAddReplica(randomComponent)) {
					System.out.println("Replicated component");
					Architecture candidateArch = (Architecture) currentBestArch
							.clone();
					candidateArch.addReplica(randomComponent);
					if (candidateArch.computeSystemReliability() >= reliablityReq) {
						if (candidateArch.getUtility(candidateArch) > currentBestUtility) {
							System.out
									.println("New architecture has better utility than previous candidate");
							currentBestArch = candidateArch;
							currentBestUtility = candidateArch
									.getUtility(candidateArch);
							currentBestArch.show();
						} else {
							System.out
									.println("New architecture does not have better utility than previous candidate");
						}
					} else {
						System.out
								.println("New architecture does not meet reliability requirement");
					}
				} else {
					System.out.println("Component cannot be replicated");

					// if possible remove replica to see if it might yield a
					// better utility
					if (currentBestArch.canRemoveReplica(randomComponent)) {
						System.out
								.println("Removed a replica of the component");
						Architecture candidateArch = (Architecture) currentBestArch
								.clone();
						candidateArch.removeReplica(randomComponent);
						if (candidateArch.computeSystemReliability() >= reliablityReq) {
							if (candidateArch.getUtility(candidateArch) > currentBestUtility) {
								System.out
										.println("New architecture has better utility than previous candidate");
								currentBestArch = candidateArch;
								currentBestUtility = candidateArch
										.getUtility(candidateArch);
								currentBestArch.show();
							} else {
								System.out
										.println("New architecture does not have better utility than previous candidate");
							}
						} else {
							System.out
									.println("New architecture does not meet reliability requirement");
						}
					} else {
						System.out
								.println("Component replica cannot be removed");
					}
				}

			} catch (Exception ex) {
				throw ex;
			}
		}
		System.out
				.println("The best architecture that can be found after running RESISTER is...");
		currentBestArch.show();
	}

	public int getNextMostUnReliableComponent(ArrayList explored) {
		int most_unreliable_comp = 0;
		double temp_rel = 1.5;
		for (int i = 0; i < no_of_components; i++) {
			if (component_rel[i] <= temp_rel && !explored.contains(i)) {
				most_unreliable_comp = i;
				temp_rel = component_rel[i];
			}
		}
		return most_unreliable_comp;
	}

	public boolean isInItsOwnProcess(int component_Id) {
		boolean inOwnProcess = true;
		int process_id = -1;
		for (int j = 0; j < no_of_processes; j++) {
			if (deployment[component_Id][j] == true) {
				process_id = j;
				break;
			}
		}
		int componentCount = 0;
		for (int i = 0; i < no_of_components; i++) {
			if (deployment[i][process_id] == true) {
				componentCount++;
			}
		}
		if (componentCount > 1)
			inOwnProcess = false;
		if (process_id == -1) {
			throw new RuntimeException("No such component has been deployed");
		}
		return inOwnProcess;
	}

	public boolean isProcessVacant(int processId) {
		return processUsage[processId] == 0;
	}

	public int getVacantProcess() {
		int process_id = -1;
		boolean isVacant = true;
		for (int j = 0; j < no_of_processes; j++) {
			isVacant = true;
			for (int i = 0; i < no_of_components; i++) {
				if (deployment[i][j] == true) {
					isVacant = false;
					break;
				}
			}
			if (isVacant) {
				process_id = j;
				break;
			}
		}
		return process_id;
	}

	public void show() {
		// display the deployment architecture
		System.out
				.println("===================================================");
		System.out.println("Deployment :");
		for (int j = 0; j < no_of_processes; j++) {
			for (int i = 0; i < no_of_components; i++) {
				if (deployment[i][j] == true)
					System.out.print("1 ");
				else
					System.out.print("0 ");
			}
			System.out.println("");
		}
		System.out.println("Replication :");
		for (int i = 0; i < no_of_components; i++) {
			System.out.print(replicas[i]);
			System.out.print(" ");
		}
		System.out.println("");
		// System.out.println("Process usage :");
		// for (int i = 0; i < no_of_processes; i++) {
		// System.out.println(processUsage[i]);
		// // System.out.print(" ");
		// }
		// System.out.println("");
		// Compute current reliability
		double rel = computeSystemReliability();
		double mem = computeSystemMemoryUtilization();
		double cpu = computeSystemCPUUtilization();
		System.out.println("System Reliability : " + rel);
		System.out.println("System memory usage : " + mem);
		System.out.println("System CPU usage : " + cpu);
		System.out.println("Overall utility : "
				+ computeOverallUtility(rel, mem, cpu));
		System.out
				.println("===================================================");
	}

	private void computeEffectiveReliability() {
		for (int j = 0; j < no_of_processes; j++) {
			double tempRel = 1;
			for (int i = 0; i < no_of_components; i++) {
				if (deployment[i][j] == true) {
					tempRel = tempRel * component_rel[i];
				}
			}
			for (int i = 0; i < no_of_components; i++) {
				if (deployment[i][j] == true) {
					if (replicas[i] == 0) {
						component_rel_eff[i] = tempRel;
					} else {
						component_rel_eff[i] = 1 - Math.pow(
								(1 - component_rel[i]), 1 + replicas[i]);
					}
				}
			}
		}
	}

	public double computeSystemReliability() {
		computeEffectiveReliability();
		if (opProfile == null)
			throw new NullPointerException("No operational profile specified");
		double[][] M = new double[no_of_components + 1][no_of_components + 1];
		for (int i = 0; i < no_of_components; i++) {
			for (int j = 0; j < no_of_components; j++) {
				M[i][j] = opProfile[i][j] * component_rel_eff[i];
			}
			M[i][no_of_components] = opProfile[i][no_of_components]
					* component_rel_eff[i];
		}

		Matrix matrix_M = new Matrix(M);
		// Matrix matrix_I = Matrix.identity(5, 5);
		Matrix matrix_I = Matrix.identity(no_of_components + 1,
				no_of_components + 1);
		Matrix matrix_I_M = matrix_I.minus(matrix_M);
		Matrix matrix_E = matrix_I_M.getMatrix(0, no_of_components - 1, 1,
				no_of_components);
		double det_E = matrix_E.det();
		double det_I_M = matrix_I_M.det();
		double final_State_Rel = component_rel_eff[0]; // assuming final state
		// is same as first state
		double system_Rel = (Math.pow(-1, opProfile.length + 1))
				* final_State_Rel * det_E / det_I_M;
		return system_Rel * 100;
	}

	private double computeSystemMemoryUtilization() {
		double memoryUsage = 0;
		// compute the memory usage for the processes
		for (int i = 0; i < no_of_processes; i++)
			if (processUsage[i] != 0)
				memoryUsage += processMemRequirement;
		// compute the memory usage for the components and their replicas
		for (int i = 0; i < no_of_components; i++) {
			memoryUsage += processMemRequirement * replicas[i];
			memoryUsage += component_mem[i] * (1 + replicas[i]);
		}
		return (memoryUsage * 100) / totalMemAvailable;
	}

	private double computeSystemCPUUtilization() {
		double cpuUsage = 0;
		// compute the CPU usage for the processes
		for (int i = 0; i < no_of_processes; i++)
			if (processUsage[i] != 0)
				cpuUsage += processCPURequirement;
		// compute the CPU usage for the components and their replicas
		for (int i = 0; i < no_of_components; i++) {
			cpuUsage += processCPURequirement * replicas[i];
			cpuUsage += component_cpu[i] * (1 + replicas[i]);
		}
		return (cpuUsage * 100) / totalCPUAvailable;
	}

	private double computeOverallUtility(double rel, double mem, double cpu) {
		return Helper.computeUtility(rel, mem, cpu);
	}

	public Object clone() {
		try {
			Architecture newArch = (Architecture) super.clone();
			newArch.component_rel = (double[]) component_rel.clone();
			newArch.replicas = (int[]) replicas.clone();
			newArch.maxReplicas = (int[]) maxReplicas.clone();
			newArch.deployment = (boolean[][]) deployment.clone();
			for (int i = 0; i < deployment.length; i++) {
				newArch.deployment[i] = deployment[i].clone();
			}
			newArch.component_rel_eff = (double[]) component_rel_eff.clone();
			newArch.opProfile = (double[][]) opProfile.clone();
			for (int i = 0; i < opProfile.length; i++) {
				newArch.opProfile[i] = opProfile[i].clone();
			}
			newArch.component_mem = component_mem.clone();
			newArch.component_cpu = component_cpu.clone();
			newArch.processUsage = processUsage.clone();
			return newArch;
		} catch (CloneNotSupportedException ex) {
			throw new Error("Clone not supported");
		}
	}

	public double getUtility(Architecture arch) {
		double rel = arch.computeSystemReliability();
		double mem = arch.computeSystemMemoryUtilization();
		double cpu = arch.computeSystemCPUUtilization();
		return arch.computeOverallUtility(rel, mem, cpu);
	}

}